@@ -59,11 +59,6 @@ enum {
 	N_SIGNALS
 };
 
-static void links_page_num_func				(GtkTreeViewColumn *tree_column,
-							 GtkCellRenderer   *cell,
-							 GtkTreeModel      *tree_model,
-							 GtkTreeIter       *iter,
-							 EvSidebarLinks    *sidebar_links);
 static void update_page_callback 			(EvPageCache       *page_cache,
 							 gint               current_page,
 						         EvSidebarLinks    *sidebar_links);
@@ -166,21 +161,40 @@ ev_sidebar_links_dispose (GObject *object)
 		sidebar->priv->page_cache = NULL;
 	}
 
-
 	G_OBJECT_CLASS (ev_sidebar_links_parent_class)->dispose (object);
 }
 
+static void
+ev_sidebar_links_map (GtkWidget *widget)
+{
+	EvSidebarLinks *links;
+
+	links = EV_SIDEBAR_LINKS (widget);
+
+	GTK_WIDGET_CLASS (ev_sidebar_links_parent_class)->map (widget);
+
+	if (links->priv->model) {
+		update_page_callback (links->priv->page_cache,
+				      ev_page_cache_get_current_page (links->priv->page_cache),
+				      links);
+	}
+}
+
 static void
 ev_sidebar_links_class_init (EvSidebarLinksClass *ev_sidebar_links_class)
 {
-	GObjectClass *g_object_class;
+	GObjectClass   *g_object_class;
+	GtkWidgetClass *widget_class;
 
 	g_object_class = G_OBJECT_CLASS (ev_sidebar_links_class);
+	widget_class = GTK_WIDGET_CLASS (ev_sidebar_links_class);
 
 	g_object_class->set_property = ev_sidebar_links_set_property;
 	g_object_class->get_property = ev_sidebar_links_get_property;
 	g_object_class->dispose = ev_sidebar_links_dispose;
 
+	widget_class->map = ev_sidebar_links_map;
+
 	signals[LINK_ACTIVATED] = g_signal_new ("link-activated",
 			 G_TYPE_FROM_CLASS (g_object_class),
 		         G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
@@ -211,8 +225,6 @@ selection_changed_callback (GtkTreeSelection   *selection,
 	GtkTreeModel *model;
 	GtkTreeIter iter;
 
-	g_return_if_fail (EV_IS_SIDEBAR_LINKS (ev_sidebar_links));
-
 	document = EV_DOCUMENT (ev_sidebar_links->priv->document);
 	g_return_if_fail (ev_sidebar_links->priv->document != NULL);
 
@@ -247,7 +259,8 @@ create_loading_model (void)
 	retval = (GtkTreeModel *)gtk_list_store_new (EV_DOCUMENT_LINKS_COLUMN_NUM_COLUMNS,
 						     G_TYPE_STRING,
 						     G_TYPE_OBJECT,
-						     G_TYPE_BOOLEAN);
+						     G_TYPE_BOOLEAN,
+						     G_TYPE_STRING);
 
 	gtk_list_store_append (GTK_LIST_STORE (retval), &iter);
 	markup = g_strdup_printf ("<span size=\"larger\" style=\"italic\">%s</span>", _("Loading..."));
@@ -261,27 +274,6 @@ create_loading_model (void)
 	return retval;
 }
 
-static gint
-get_page_from_link (EvLink *link)
-{
-	EvLinkAction *action;
-	EvLinkDest *dest;
-
-	action = ev_link_get_action (link);
-	if (!action)
-		return -1;
-
-	if (ev_link_action_get_action_type (action) !=
-	    EV_LINK_ACTION_TYPE_GOTO_DEST)
-		return -1;
-
-	dest = ev_link_action_get_dest (action);
-	if (dest)
-		return ev_link_dest_get_page (dest);
-		
-	return -1;
-}
-
 static void
 print_section_cb (GtkWidget *menuitem, EvSidebarLinks *sidebar)
 {
@@ -304,7 +296,7 @@ print_section_cb (GtkWidget *menuitem, EvSidebarLinks *sidebar)
 		if (!link)
 			return;
 
-		first_page = get_page_from_link (link);
+		first_page = ev_link_get_page (link);
 		if (first_page == -1) {
 			g_object_unref (link);
 			return;
@@ -319,7 +311,7 @@ print_section_cb (GtkWidget *menuitem, EvSidebarLinks *sidebar)
 					    -1);
 
 			if (link) {
-				last_page = get_page_from_link (link);
+				last_page = ev_link_get_page (link);
 				g_object_unref (link);
 			}
 		} else {
@@ -419,7 +411,7 @@ ev_sidebar_links_construct (EvSidebarLinks *ev_sidebar_links)
 	loading_model = create_loading_model ();
 	priv->tree_view = gtk_tree_view_new_with_model (loading_model);
 	g_object_unref (loading_model);
-
+	
 	selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (priv->tree_view));
 	gtk_tree_selection_set_mode (selection, GTK_SELECTION_NONE);
 	gtk_tree_view_set_headers_visible (GTK_TREE_VIEW (priv->tree_view), FALSE);
@@ -444,9 +436,10 @@ ev_sidebar_links_construct (EvSidebarLinks *ev_sidebar_links)
 	
 	renderer = gtk_cell_renderer_text_new ();
 	gtk_tree_view_column_pack_end (GTK_TREE_VIEW_COLUMN (column), renderer, FALSE);
-	gtk_tree_view_column_set_cell_data_func (GTK_TREE_VIEW_COLUMN (column), renderer,
-						 (GtkTreeCellDataFunc) links_page_num_func,
-						 ev_sidebar_links, NULL);
+	gtk_tree_view_column_set_attributes (GTK_TREE_VIEW_COLUMN (column), renderer,
+					     "text", EV_DOCUMENT_LINKS_COLUMN_PAGE_LABEL,
+					     NULL);
+	g_object_set (G_OBJECT (renderer), "style", PANGO_STYLE_ITALIC, NULL);
 
 	g_signal_connect (GTK_TREE_VIEW (priv->tree_view),
 			  "button_press_event",
@@ -466,51 +459,38 @@ ev_sidebar_links_init (EvSidebarLinks *ev_sidebar_links)
 	ev_sidebar_links_construct (ev_sidebar_links);
 }
 
-static void
-links_page_num_func (GtkTreeViewColumn *tree_column,
-		     GtkCellRenderer   *cell,
-		     GtkTreeModel      *tree_model,
-		     GtkTreeIter       *iter,
-		     EvSidebarLinks    *sidebar_links)
+static gboolean
+fill_page_labels (GtkTreeModel *tree_model,
+                  GtkTreePath *path,
+                  GtkTreeIter *iter,
+    	          EvSidebarLinks    *sidebar_links)
 {
 	EvLink *link;
 	gint page;
+	gchar *page_label;
 
 	gtk_tree_model_get (tree_model, iter,
 			    EV_DOCUMENT_LINKS_COLUMN_LINK, &link,
 			    -1);
 
-	if (!link) {
-		g_object_set (cell,
-			      "visible", FALSE,
-			      NULL);
-		return;
-	}
+	if (!link)
+		return FALSE;
 
-	page = get_page_from_link (link);
+	page = ev_link_get_page (link);
 
-	if (page >= 0) {
-		gchar *page_label;
-		gchar *page_string;
+	if (page < 0) 
+		return FALSE;
+	
+	page_label = ev_page_cache_get_page_label (sidebar_links->priv->page_cache,
+						   page);
+	gtk_tree_store_set (GTK_TREE_STORE (tree_model), iter,
+			    EV_DOCUMENT_LINKS_COLUMN_PAGE_LABEL, page_label, 
+			    -1);
 
-		page_label = ev_page_cache_get_page_label (sidebar_links->priv->page_cache,
-							   page);
-		page_string = g_markup_printf_escaped ("<i>%s</i>", page_label);
-
-		g_object_set (cell,
- 			      "markup", page_string,
-			      "visible", TRUE,
-			      NULL);
-
-		g_free (page_label);
-		g_free (page_string);
-	} else {
-		g_object_set (cell,
-			      "visible", FALSE,
-			      NULL);
-	}
+	g_free (page_label);
 
 	g_object_unref (link);
+	return FALSE;
 }
 
 /* Public Functions */
@@ -542,18 +522,16 @@ update_page_callback_foreach (GtkTreeModel *model,
 		int current_page;
 		int dest_page;
 
-		dest_page = get_page_from_link (link);
+		dest_page = ev_link_get_page (link);
 		g_object_unref (link);
 		
 		current_page = ev_page_cache_get_current_page (sidebar_links->priv->page_cache);
 			 
 		if (dest_page == current_page) {
-			GtkTreeSelection *selection;
-
-			selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (sidebar_links->priv->tree_view));
 			gtk_tree_view_expand_to_path (GTK_TREE_VIEW (sidebar_links->priv->tree_view),
 						      path);
-			gtk_tree_selection_select_path (selection, path);
+			gtk_tree_view_set_cursor (GTK_TREE_VIEW (sidebar_links->priv->tree_view),
+						  path, NULL, FALSE);
 			
 			return TRUE;
 		}
@@ -571,6 +549,10 @@ update_page_callback (EvPageCache    *page_cache,
 	GtkTreeModel *model;
 	GtkTreeIter iter;
 
+	/* Widget is not currently visible */
+	if (!GTK_WIDGET_MAPPED (sidebar_links))
+		return;
+	
 	selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (sidebar_links->priv->tree_view));
 
 	if (gtk_tree_selection_get_selected (selection, &model, &iter)) {
@@ -580,13 +562,11 @@ update_page_callback (EvPageCache    *page_cache,
 				    EV_DOCUMENT_LINKS_COLUMN_LINK, &link,
 				    -1);
 		if (link) {
-			gint current_page;
 			gint dest_page;
 
-			dest_page = get_page_from_link (link);
+			dest_page = ev_link_get_page (link);
 			g_object_unref (link);
 			
-			current_page = ev_page_cache_get_current_page (sidebar_links->priv->page_cache);
 			if (dest_page == current_page)
 				return;
 		}
@@ -604,24 +584,21 @@ update_page_callback (EvPageCache    *page_cache,
 				update_page_callback_foreach,
 				sidebar_links);
 	
-
 	g_signal_handler_unblock (selection, sidebar_links->priv->selection_id);
 	g_signal_handler_unblock (sidebar_links->priv->tree_view, sidebar_links->priv->row_activated_id);
 }
 
 static void 
-row_activated_callback 			(GtkTreeView *treeview,
-                                         GtkTreePath *arg1,
-                                         GtkTreeViewColumn *arg2,
-                                         gpointer user_data)
+row_activated_callback (GtkTreeView       *treeview,
+			GtkTreePath       *arg1,
+			GtkTreeViewColumn *arg2,
+			gpointer           user_data)
 {	
 	if (gtk_tree_view_row_expanded (GTK_TREE_VIEW (treeview), arg1)) {
-		    gtk_tree_view_collapse_row (GTK_TREE_VIEW (treeview), arg1);
+		gtk_tree_view_collapse_row (GTK_TREE_VIEW (treeview), arg1);
 	} else {
-		    gtk_tree_view_expand_row (GTK_TREE_VIEW (treeview), arg1, FALSE);
+		gtk_tree_view_expand_row (GTK_TREE_VIEW (treeview), arg1, FALSE);
 	}
-	
-        return;
 }
 
 static void
@@ -637,7 +614,7 @@ expand_open_links (GtkTreeView *tree_view, GtkTreeModel *model, GtkTreeIter *par
 					    -1);
 			if (expand) {
 				GtkTreePath *path;
-
+				
 				path = gtk_tree_model_get_path (model, &iter);
 				gtk_tree_view_expand_row (tree_view, path, FALSE);
 				gtk_tree_path_free (path);
@@ -659,6 +636,8 @@ job_finished_callback (EvJobLinks     *job,
 	
 	priv->model = job->model;
 	g_object_notify (G_OBJECT (sidebar_links), "model");
+	
+	gtk_tree_model_foreach (priv->model, (GtkTreeModelForeachFunc)fill_page_labels, sidebar_links);
 
 	gtk_tree_view_set_model (GTK_TREE_VIEW (priv->tree_view), job->model);
 	
